#  Copyright (c) New Cloud Technologies, Ltd. 2013-2022.
#  Author: Vitaly Isaev <vitaly.isaev@myoffice.team>
#  License: https://github.com/newcloudtechnologies/rbfopt-go/blob/master/LICENSE

import pathlib
from typing import List

import numpy as np
import pandas as pd

from rbfoptgo.common import Cost, ParameterValue
from rbfoptgo.client import Client
from rbfoptgo.config import Config
from rbfoptgo.report import Report
from rbfoptgo import names


class Evaluator:
    """
    Evaluator is responsible for cost function estimation. It performs HTTP calls to the Golang part of a library.
    """
    __config: Config
    __client: Client
    __parameter_names: List[str]
    __evaluations: []
    __root_dir: pathlib.Path
    __report: Report
    __iterations: int

    def __init__(self, config: Config, client: Client, parameter_names: List[str], root_dir: pathlib.Path):
        self.__config = config
        self.__client = client
        self.__parameter_names = parameter_names
        self.__evaluations = []
        self.__root_dir = root_dir
        self.__iterations = 0

    def __np_array_to_parameter_values(self, raw_values: np.ndarray) -> List[ParameterValue]:
        parameter_values = []
        for i, raw_value in enumerate(raw_values):
            parameter_values.append(
                ParameterValue(name=self.__parameter_names[i], value=int(raw_value)),
            )
        return parameter_values

    def estimate_cost(self, raw_values: np.ndarray) -> Cost:
        """
        :param raw_values: vector of cost function arguments
        :return: cost function particular value
        """
        self.__iterations += 1

        parameter_values = self.__np_array_to_parameter_values(raw_values)
        cost, invalid_parameter_combination = self.__client.estimate_cost(parameter_values)

        # store evaluation result for the future use
        entry = dict(zip(self.__parameter_names, raw_values))
        entry[names.Iteration] = self.__iterations
        entry[names.Cost] = cost
        entry[names.InvalidParameterCombination] = invalid_parameter_combination

        self.__evaluations.append(entry)

        return cost

    def register_report(
            self,
            cost: Cost,
            optimum: np.ndarray,
            iterations: int,
            evaluations: int,
            fast_evaluations: int,
    ):
        """
        Registers final report.
        :param cost: minimal value of a cost function (the optimum)
        :param optimum: the values of cost function arguments corresponding to the optimum
        :param iterations: number of iterations performed
        :param evaluations: number of evaluations performed
        :param fast_evaluations: number of fast_evaluations
        :return: None
        """
        report = Report(
            bounds=self.__config.rbfopt.parameters,
            cost=cost,
            optimum=self.__np_array_to_parameter_values(optimum),
            iterations=iterations,
            evaluations=evaluations,
            fast_evaluations=fast_evaluations,
        )

        self.__client.register_report(report)
        self.__report = report

    def dump(self) -> (pd.DataFrame, Report):
        """
        Returns the list of performed cost function evaluations.
        :return: DataFrame + json-serializable Report compatible with Golang library
        """
        # dump history of evaluations for future usage
        evaluations = pd.DataFrame(self.__evaluations)
        file_path = self.__root_dir.joinpath("evaluations.csv")
        evaluations.to_csv(file_path, header=True, index=False)

        # dump report for future usage
        file_path = self.__root_dir.joinpath("report.json")
        self.__report.save_to_file(file_path)

        return evaluations, self.__report



#  Copyright (c) New Cloud Technologies, Ltd. 2013-2022.
#  Author: Vitaly Isaev <vitaly.isaev@myoffice.team>
#  License: https://github.com/newcloudtechnologies/rbfopt-go/blob/master/LICENSE

import glob
import pathlib

from rbfoptgo.plot_debug import run


def test_plot():
    """
    Integration test to estimate approximate Python coverage
    """
    # Directory with evaluations left after Go tests
    debug_dir = glob.glob("/tmp/rbfopt*")[0]
    run(pathlib.Path(debug_dir))



#  Copyright (c) New Cloud Technologies, Ltd. 2013-2022.
#  Author: Vitaly Isaev <vitaly.isaev@myoffice.team>
#  License: https://github.com/newcloudtechnologies/rbfopt-go/blob/master/LICENSE

from http import HTTPStatus
from typing import List
from urllib.parse import urljoin

import jsons
import requests

from rbfoptgo.common import Cost, ParameterValue
from rbfoptgo.report import Report
from rbfoptgo import names


class Client:
    """
    HTTP client to the Go part of library
    """
    url_head: str
    session: requests.Session

    def __init__(self, endpoint: str):
        self.url_head = f'http://{endpoint}'
        self.session = requests.Session()

    def estimate_cost(self, parameter_values: List[ParameterValue]) -> (Cost, bool):
        """
        Requests a particular cost function value for a given parameters
        :param parameter_values: a vector of parameters
        :return: 1. The value of a cost function
               2. Sign that optimizer gave a vector of parameters that was considered as non-optimal
        """
        print(f"request '{parameter_values}'")

        payload = dict(parameter_values=parameter_values)
        response = self.session.get(
            urljoin(self.url_head, 'estimate_cost'),
            json=jsons.dump(payload),
        )

        print(f"response code={response.status_code} body={response.json()}")

        if response.status_code != HTTPStatus.OK:
            raise ValueError(f'invalid status code {response.status_code}')

        return response.json()[names.Cost], response.json()[names.InvalidParameterCombination]

    def register_report(self, report: Report):
        """
        Posts the report to the Go sides
        :param report: optimizer report itself
        :return:
        """
        print(f"request '{report}'")

        payload = dict(report=report)
        response = self.session.post(
            urljoin(self.url_head, 'register_report'),
            json=jsons.dump(payload),
        )

        print(f"response code={response.status_code} response={response}")

        if response.status_code != HTTPStatus.OK:
            raise ValueError(f'invalid status code {response.status_code}')



#  Copyright (c) New Cloud Technologies, Ltd. 2013-2022.
#  Author: Vitaly Isaev <vitaly.isaev@myoffice.team>
#  License: https://github.com/newcloudtechnologies/rbfopt-go/blob/master/LICENSE
import json
import os
import pathlib
from dataclasses import dataclass
from datetime import datetime
from time import mktime
from enum import Enum
from typing import List, Any, Dict

import numpy as np


@dataclass
class Bound:
    """
    Describes the area of variation for a cost function parameter
    """
    left: int
    right: int

    @staticmethod
    def from_dict(obj: Any) -> 'Bound':
        """
        Constructs object from an arbitrary dictionary
        :param obj: Dictionary with parameter values
        :return: an object of desired type
        """
        _left = int(obj.get("left"))
        _right = int(obj.get("right"))
        return Bound(_left, _right)


@dataclass
class Parameter:
    """
    An arbitrary parameter of a cost function
    """
    bound: Bound
    name: str

    @staticmethod
    def from_dict(obj: Any) -> 'Parameter':
        """
        Constructs object from an arbitrary dictionary
        :param obj: Dictionary with parameter values
        :return: an object of desired type
        """
        _bound = Bound.from_dict(obj.get("bound"))
        _name = str(obj.get("name"))
        return Parameter(_bound, _name)


class InvalidParameterCombinationRenderPolicy(Enum):
    """
    Describes the behavior of the plot renderer when the computed cost function value
    corresponds to ErrInvalidParameterCombinationCost.
    """
    omit = 1
    assign_closest_valid_value = 2


@dataclass
class PlotConfig:
    """
    Configuration of plotting config.
    """
    scatter_plot_policy: InvalidParameterCombinationRenderPolicy
    heatmap_render_policy: InvalidParameterCombinationRenderPolicy

    @staticmethod
    def from_dict(obj: Any) -> 'PlotConfig':
        """
        Constructs object from an arbitrary dictionary
        :param obj: Dictionary with parameter values
        :return: an object of desired type
        """
        _scatter_plot_policy = InvalidParameterCombinationRenderPolicy[str(obj.get("scatter_plot_policy"))]
        _heatmap_render_policy = InvalidParameterCombinationRenderPolicy[str(obj.get("heatmap_render_policy"))]
        return PlotConfig(_scatter_plot_policy, _heatmap_render_policy)


@dataclass
class RBFOptConfig:
    """
    Configuration of RBFOpt library.
    """
    parameters: List[Parameter]
    max_evaluations: int
    max_iterations: int
    init_strategy: str
    invalid_parameter_combination_cost: int

    @staticmethod
    def from_dict(obj: Any) -> 'RBFOptConfig':
        """
        Constructs object from an arbitrary dictionary
        :param obj: Dictionary with parameter values
        :return: an object of desired type
        """
        _parameters = [Parameter.from_dict(y) for y in obj.get("parameters")]
        _max_evaluations = int(obj.get("max_evaluations"))
        _max_iterations = int(obj.get("max_iterations"))
        _init_strategy = str(obj.get("init_strategy"))
        _invalid_parameter_combination_cost = int(obj.get("invalid_parameter_combination_cost"))
        return RBFOptConfig(_parameters, _max_evaluations, _max_iterations, _init_strategy,
                            _invalid_parameter_combination_cost)

    @property
    def var_names(self) -> List[str]:
        """
        Returns parameter names suitable for RBFOpt
        :return:  list of parameter namse
        """
        var_names = []
        for param in self.parameters:
            var_names.append(param.name)

        return var_names

    @property
    def user_black_box(self) -> Dict:
        """
        Returns the area of the function scope where to perform optimization.
        :return:  list of parameter namse
        """
        dimensions = len(self.parameters)
        var_lower = np.zeros(shape=(dimensions,))
        var_upper = np.zeros(shape=(dimensions,))
        var_types = np.chararray(shape=(dimensions,))

        for i, param in enumerate(self.parameters):
            var_lower[i] = param.bound.left
            var_upper[i] = param.bound.right
            var_types[i] = 'I'

        return dict(
            dimension=dimensions,
            var_lower=var_lower,
            var_upper=var_upper,
            var_type=var_types,
        )

    @property
    def settings(self) -> Dict:
        """
        Returns dictionary with RBFOpt settings.
        :return: Dict with settings.
        """
        return dict(
            max_evaluations=self.max_evaluations,
            max_iterations=self.max_iterations,
            rand_seed=int(mktime(datetime.now().timetuple())),
            init_strategy=self.init_strategy,
        )


@dataclass
class Config:
    """
    A configuration for the Python part of rbfopt-go
    """
    root_dir: os.PathLike
    endpoint: str
    rbfopt: RBFOptConfig
    plot: PlotConfig

    @staticmethod
    def from_dict(obj: Any) -> 'Config':
        """
        Constructs object from an arbitrary dictionary
        :param obj: Dictionary with parameter values
        :return: an object of desired type
        """
        _root_dir = pathlib.Path(str(obj.get("root_dir")))
        _endpoint = str(obj.get("endpoint"))
        _rbfopt = RBFOptConfig.from_dict(obj.get("rbfopt"))
        _plot = PlotConfig.from_dict(obj.get("plot"))
        return Config(_root_dir, _endpoint, _rbfopt, _plot)

    @staticmethod
    def from_file(path: os.PathLike) -> 'Config':
        """
        Loads configuration from file
        :param path: full path to config file
        :return:
        """
        with open(path, "r") as f:
            data = json.load(f)
            return Config.from_dict(data)
